import { CountMinSketch } from './sketch';
import { ON_REMOVE, ON_EVICT, TRIGGER_REMOVE, EVICT } from '../symbols';
import { RemovalReason } from '../removal-reason';
import { CacheNode } from '../cache-node';
import { AbstractCache } from '../abstract';
const percentInMain = 0.99;
const percentProtected = 0.8;
const percentOverflow = 0.01;
const DATA = Symbol('boundedData');
/**
 * Bounded cache implementation using W-TinyLFU to keep track of data.
 *
 * See https://arxiv.org/pdf/1512.00727.pdf for details about TinyLFU and
 * the W-TinyLFU optimization.
 */
export class BoundedCache extends AbstractCache {
    constructor(options) {
        super();
        const maxMain = Math.floor(percentInMain * options.maxSize);
        /*
         * For weighted caches use an initial sketch size of 256. It will
         * grow when the size of the cache approaches that size.
         *
         * Otherwise set it to a minimum of 128 or the maximum requested size
         * of the graph.
         */
        const sketchWidth = options.weigher ? 256 : Math.max(options.maxSize, 128);
        this[DATA] = {
            maxSize: options.weigher ? -1 : options.maxSize,
            removalListener: options.removalListener || null,
            weigher: options.weigher || null,
            weightedMaxSize: options.maxSize,
            weightedSize: 0,
            sketch: CountMinSketch.uint8(sketchWidth, 4),
            sketchGrowLimit: sketchWidth,
            values: new Map(),
            // Tracking of the window cache, around 1% of the total cache
            window: {
                head: new BoundedNode(null, null),
                size: 0,
                maxSize: options.maxSize - maxMain
            },
            // SLRU protected segment, 80% * 99% of the total cache
            protected: {
                head: new BoundedNode(null, null),
                size: 0,
                maxSize: Math.floor(maxMain * percentProtected)
            },
            // SLRU probation segment, 20% * 99% of the total cache
            probation: {
                head: new BoundedNode(null, null),
            },
            // Timeout used to schedule evictions
            evictionTimeout: 0,
            // The maximum size we can temporarily be grow before an eviction is forced
            forceEvictionLimit: options.maxSize + Math.max(Math.floor(options.maxSize * percentOverflow), 5),
            // The time to wait before an eviction is triggered by a set
            evictionInterval: 5000
        };
    }
    /**
     * Get the maximum size this cache can be.
     */
    get maxSize() {
        return this[DATA].maxSize;
    }
    /**
     * Get the current size of the cache.
     */
    get size() {
        return this[DATA].values.size;
    }
    /**
     * Get the weighted size of all items in the cache.
     */
    get weightedSize() {
        return this[DATA].weightedSize;
    }
    /**
     * Cache and associate a value with the given key.
     */
    set(key, value) {
        const data = this[DATA];
        const old = data.values.get(key);
        // Create a node and add it to the backing map
        const node = new BoundedNode(key, value);
        data.values.set(key, node);
        if (data.weigher) {
            node.weight = data.weigher(key, value);
        }
        // Update our weight
        data.weightedSize += node.weight;
        if (old) {
            // Remove the old node
            old.remove();
            // Ajudst weight
            data.weightedSize -= old.weight;
            // Update weights of where the node belonged
            switch (old.location) {
                case 1 /* PROTECTED */:
                    // Node was protected, reduce the size
                    data.protected.size -= old.weight;
                    break;
                case 0 /* WINDOW */:
                    // Node was in window, reduce window size
                    data.window.size -= old.weight;
                    break;
            }
        }
        // Check if we reached the grow limit of the sketch
        if (data.weigher && data.values.size >= data.sketchGrowLimit) {
            const sketchWidth = data.values.size * 2;
            data.sketch = CountMinSketch.uint8(sketchWidth, 4);
            data.sketchGrowLimit = sketchWidth;
        }
        // Append the new node to the window space
        node.append(data.window.head);
        data.window.size += node.weight;
        // Register access to the key
        data.sketch.update(node.hashCode);
        // Schedule eviction
        if (data.weightedSize >= data.forceEvictionLimit) {
            this[EVICT]();
        }
        else if (!data.evictionTimeout) {
            data.evictionTimeout = setTimeout(() => this[EVICT](), data.evictionInterval);
        }
        // Return the value we replaced
        if (old) {
            this[TRIGGER_REMOVE](key, old.value, RemovalReason.REPLACED);
            return old.value;
        }
        else {
            return null;
        }
    }
    /**
     * Get a previously cached value.
     */
    getIfPresent(key) {
        const data = this[DATA];
        const node = data.values.get(key);
        if (!node) {
            // This value does not exist in the cache
            return null;
        }
        // Register access to the key
        data.sketch.update(node.hashCode);
        switch (node.location) {
            case 0 /* WINDOW */:
                // In window cache, mark as most recently used
                node.move(data.window.head);
                break;
            case 2 /* PROBATION */:
                // In SLRU probation segment, move to protected
                node.location = 1 /* PROTECTED */;
                node.move(data.protected.head);
                // Plenty of room, keep track of the size
                data.protected.size += node.weight;
                while (data.protected.size > data.protected.maxSize) {
                    /*
                        * There is now too many nodes in the protected segment
                        * so demote the least recently used.
                        */
                    const lru = data.protected.head.next;
                    lru.location = 2 /* PROBATION */;
                    lru.move(data.probation.head);
                    data.protected.size -= lru.weight;
                }
                break;
            case 1 /* PROTECTED */:
                // SLRU protected segment, mark as most recently used
                node.move(data.protected.head);
                break;
        }
        return node.value;
    }
    peek(key) {
        const data = this[DATA];
        const node = data.values.get(key);
        return node ? node.value : null;
    }
    /**
     * Delete any value associated with the given key from the cache.
     */
    delete(key) {
        const data = this[DATA];
        const node = data.values.get(key);
        if (node) {
            // Remove the node from its current list
            node.remove();
            switch (node.location) {
                case 1 /* PROTECTED */:
                    // Node was protected, reduce the size
                    data.protected.size -= node.weight;
                    break;
                case 0 /* WINDOW */:
                    // Node was in window, reduce window size
                    data.window.size -= node.weight;
                    break;
            }
            // Reduce overall weight
            data.weightedSize -= node.weight;
            // Remove from main value storage
            data.values.delete(key);
            this[TRIGGER_REMOVE](key, node.value, RemovalReason.EXPLICIT);
            if (!data.evictionTimeout) {
                data.evictionTimeout = setTimeout(() => this[EVICT](), data.evictionInterval);
            }
            return node.value;
        }
        return null;
    }
    /**
     * Check if a certain value exists in the cache.
     */
    has(key) {
        const data = this[DATA];
        return data.values.has(key);
    }
    clear() {
        const data = this[DATA];
        const oldValues = data.values;
        data.values = new Map();
        for (const [key, node] of oldValues) {
            this[TRIGGER_REMOVE](key, node.value, RemovalReason.EXPLICIT);
        }
        data.weightedSize = 0;
        data.window.head.remove();
        data.window.size = 0;
        data.probation.head.remove();
        data.protected.head.remove();
        data.protected.size = 0;
        if (data.evictionTimeout) {
            clearTimeout(data.evictionTimeout);
            data.evictionTimeout = null;
        }
    }
    keys() {
        this[EVICT]();
        return Array.from(this[DATA].values.keys());
    }
    cleanUp() {
        this[EVICT]();
    }
    get metrics() {
        throw new Error('Metrics are not supported by this cache');
    }
    [TRIGGER_REMOVE](key, value, cause) {
        const data = this[DATA];
        // Trigger any extended remove listeners
        const onRemove = this[ON_REMOVE];
        if (onRemove) {
            onRemove(key, value, cause);
        }
        // Trigger the removal listener
        if (data.removalListener) {
            data.removalListener(key, value, cause);
        }
    }
    [EVICT]() {
        const data = this[DATA];
        /*
        * Evict the least recently used item in the window space to the
        * probation segment until we are below the maximum size.
        */
        let evictedToProbation = 0;
        while (data.window.size > data.window.maxSize) {
            const first = data.window.head.next;
            first.move(data.probation.head);
            first.location = 2 /* PROBATION */;
            data.window.size -= first.weight;
            evictedToProbation++;
        }
        /*
         * Evict items for real until we are below our maximum size.
         */
        while (data.weightedSize > data.weightedMaxSize) {
            const probation = data.probation.head.next;
            //const evicted = evictedToProbation == 0 ? null : data.probation.head.previous;
            const evicted = evictedToProbation === 0 ? data.probation.head : data.probation.head.previous;
            const hasProbation = probation !== data.probation.head;
            const hasEvicted = evicted !== data.probation.head;
            let toRemove;
            if (!hasProbation && !hasEvicted) {
                // TODO: Probation queue is empty, how is this handled?
                break;
            }
            else if (!hasEvicted) {
                toRemove = probation;
            }
            else if (!hasProbation) {
                toRemove = evicted;
                evictedToProbation--;
            }
            else {
                // Estimate how often the two items have been accessed
                const freqEvicted = data.sketch.estimate(evicted.hashCode);
                const freqProbation = data.sketch.estimate(probation.hashCode);
                if (freqEvicted > freqProbation) {
                    toRemove = probation;
                }
                else {
                    toRemove = evicted;
                }
                evictedToProbation--;
            }
            if (toRemove.key === null) {
                throw new Error('Cache issue, problem with removal');
            }
            data.values.delete(toRemove.key);
            toRemove.remove();
            data.weightedSize -= toRemove.weight;
            this[TRIGGER_REMOVE](toRemove.key, toRemove.value, RemovalReason.SIZE);
        }
        // Trigger the onEvict listener if one exists
        const onEvict = this[ON_EVICT];
        if (onEvict) {
            onEvict();
        }
        if (data.evictionTimeout) {
            clearTimeout(data.evictionTimeout);
            data.evictionTimeout = null;
        }
    }
}
/**
 * Node in a double-linked list.
 */
class BoundedNode extends CacheNode {
    constructor(key, value) {
        super(key, value);
        this.hashCode = key == null ? 0 : CountMinSketch.hash(key);
        this.weight = 1;
        this.location = 0 /* WINDOW */;
    }
}
//# sourceMappingURL=index.js.map