import { AbstractCache } from '../abstract';
import { RemovalReason } from '../removal-reason';
import { ON_REMOVE, ON_EVICT, TRIGGER_REMOVE, EVICT } from '../symbols';
const DATA = Symbol('boundlessData');
const EVICTION_DELAY = 5000;
/**
 * Boundless cache.
 */
export class BoundlessCache extends AbstractCache {
    constructor(options) {
        super();
        this[DATA] = {
            values: new Map(),
            removalListener: options.removalListener || null,
            evictionTimeout: null
        };
    }
    /**
     * Get the maximum size this cache can be.
     */
    get maxSize() {
        return -1;
    }
    /**
     * Get the current size of the cache.
     */
    get size() {
        return this[DATA].values.size;
    }
    get weightedSize() {
        return this.size;
    }
    /**
     * Cache and associate a value with the given key.
     */
    set(key, value) {
        const data = this[DATA];
        const old = data.values.get(key);
        // Update with the new value
        data.values.set(key, value);
        // Schedule an eviction
        if (!data.evictionTimeout) {
            data.evictionTimeout = setTimeout(() => this[EVICT](), EVICTION_DELAY);
        }
        // Return the value we replaced
        if (old !== undefined) {
            this[TRIGGER_REMOVE](key, old, RemovalReason.REPLACED);
            return old;
        }
        else {
            return null;
        }
    }
    /**
     * Get a value from this cache if it has been previously cached.
     */
    getIfPresent(key) {
        const data = this[DATA];
        const value = data.values.get(key);
        return value === undefined ? null : value;
    }
    peek(key) {
        const data = this[DATA];
        const value = data.values.get(key);
        return value === undefined ? null : value;
    }
    /**
     * Delete any value associated with the given key from the cache.
     */
    delete(key) {
        const data = this[DATA];
        const old = data.values.get(key);
        data.values.delete(key);
        if (old !== undefined) {
            // Trigger removal events
            this[TRIGGER_REMOVE](key, old, RemovalReason.EXPLICIT);
            // Queue an eviction event if one is not set
            if (!data.evictionTimeout) {
                data.evictionTimeout = setTimeout(() => this[EVICT](), EVICTION_DELAY);
            }
            return old;
        }
        else {
            return null;
        }
    }
    /**
     * Check if a certain value exists in the cache.
     */
    has(key) {
        const data = this[DATA];
        return data.values.has(key);
    }
    /**
     * Clear all of the cached data.
     */
    clear() {
        const data = this[DATA];
        const oldValues = data.values;
        // Simply replace the value map new data
        data.values = new Map();
        // Trigger removal events for all of the content in the cache
        for (const [key, value] of oldValues.entries()) {
            this[TRIGGER_REMOVE](key, value, RemovalReason.EXPLICIT);
        }
    }
    /**
     * Get all of the keys currently in the cache.
     */
    keys() {
        this[EVICT]();
        return Array.from(this[DATA].values.keys());
    }
    /**
     * Clean up.
     */
    cleanUp() {
        // Simply request eviction so extra layers can handle this
        this[EVICT]();
    }
    get metrics() {
        throw new Error('Metrics are not supported by this cache');
    }
    [TRIGGER_REMOVE](key, value, reason) {
        const data = this[DATA];
        // Trigger any extended remove listeners
        const onRemove = this[ON_REMOVE];
        if (onRemove) {
            onRemove(key, value, reason);
        }
        if (data.removalListener) {
            data.removalListener(key, value, reason);
        }
    }
    [EVICT]() {
        // Trigger the onEvict listener if one exists
        const onEvict = this[ON_EVICT];
        if (onEvict) {
            onEvict();
        }
        const data = this[DATA];
        if (data.evictionTimeout) {
            clearTimeout(data.evictionTimeout);
            data.evictionTimeout = null;
        }
    }
}
//# sourceMappingURL=index.js.map