import { AbstractCache } from '../abstract';
import { RemovalReason } from '../removal-reason';
import { TimerWheel } from './timer-wheel';
import { PARENT, ON_REMOVE, TRIGGER_REMOVE, ON_EVICT } from '../symbols';
const DATA = Symbol('expirationData');
/**
 * Wrapper for another cache that provides evictions of times based on timers.
 *
 * Currently supports expiration based on maximum age.
 */
export class ExpirationCache extends AbstractCache {
    constructor(options) {
        super();
        this[PARENT] = options.parent;
        this[DATA] = {
            maxWriteAge: options.maxWriteAge,
            maxNoReadAge: options.maxNoReadAge,
            removalListener: options.removalListener || null,
            timerWheel: new TimerWheel(keys => {
                for (const key of keys) {
                    this.delete(key);
                }
            })
        };
        // Custom onRemove handler for the parent cache
        this[PARENT][ON_REMOVE] = (key, node, reason) => {
            if (node.isExpired()) {
                reason = RemovalReason.EXPIRED;
            }
            this[DATA].timerWheel.deschedule(node);
            this[TRIGGER_REMOVE](key, node.value, reason);
        };
        // Custom eviction behaviour to advance the wheel
        this[PARENT][ON_EVICT] = () => this[DATA].timerWheel.advance();
    }
    get maxSize() {
        return this[PARENT].maxSize;
    }
    get size() {
        return this[PARENT].size;
    }
    get weightedSize() {
        return this[PARENT].weightedSize;
    }
    set(key, value) {
        const data = this[DATA];
        const timerWheel = data.timerWheel;
        const node = timerWheel.node(key, value);
        let age = null;
        if (data.maxWriteAge) {
            age = data.maxWriteAge(key, value) || 0;
        }
        else if (data.maxNoReadAge) {
            age = data.maxNoReadAge(key, value) || 0;
        }
        if (age !== null && !data.timerWheel.schedule(node, age)) {
            // Age was not accepted by wheel, delete any previous value
            return this.delete(key);
        }
        try {
            const replaced = this[PARENT].set(key, node);
            return replaced ? replaced.value : null;
        }
        catch (ex) {
            timerWheel.deschedule(node);
            throw ex;
        }
    }
    getIfPresent(key) {
        const node = this[PARENT].getIfPresent(key);
        if (node) {
            if (node.isExpired()) {
                // Check if the node is expired and return null if so
                return null;
            }
            // Reschedule if we have a maximum age between reads
            const data = this[DATA];
            if (data.maxNoReadAge) {
                const age = data.maxNoReadAge(key, node.value);
                if (!data.timerWheel.schedule(node, age)) {
                    // Age was not accepted by wheel, expire it directly
                    this.delete(key);
                }
            }
            return node.value;
        }
        return null;
    }
    peek(key) {
        const node = this[PARENT].peek(key);
        return node && !node.isExpired() ? node.value : null;
    }
    has(key) {
        const node = this[PARENT].peek(key);
        return (node && !node.isExpired()) || false;
    }
    delete(key) {
        const node = this[PARENT].delete(key);
        return node ? node.value : null;
    }
    clear() {
        return this[PARENT].clear();
    }
    keys() {
        return this[PARENT].keys();
    }
    cleanUp() {
        return this[PARENT].cleanUp();
    }
    get metrics() {
        return this[PARENT].metrics;
    }
    get [ON_EVICT]() {
        return this[PARENT][ON_EVICT];
    }
    set [ON_EVICT](listener) {
        this[PARENT][ON_EVICT] = listener;
    }
    [TRIGGER_REMOVE](key, value, reason) {
        // Trigger any extended remove listeners
        const onRemove = this[ON_REMOVE];
        if (onRemove) {
            onRemove(key, value, reason);
        }
        const data = this[DATA];
        // Trigger the removal listener
        if (data.removalListener) {
            data.removalListener(key, value, reason);
        }
    }
}
//# sourceMappingURL=index.js.map