import { Cache } from '../cache';
import { CacheSPI } from '../cache-spi';
import { Metrics } from '../metrics/metrics';
import { ON_REMOVE, ON_EVICT, TRIGGER_REMOVE, EVICT } from '../symbols';
import { Weigher } from '../weigher';
import { KeyType } from '../key-type';
import { AbstractCache } from '../abstract';
import { RemovalListener } from '../removal-listener';
declare const DATA: unique symbol;
/**
 * Options usable with a BoundedCache.
 */
export interface BoundedCacheOptions<K extends KeyType, V> {
    /**
     * The maximum size of the cache. For unweighed caches this is the maximum
     * number of entries in the cache, for weighed caches this is the maximum
     * weight of the cache.
     */
    maxSize: number;
    /**
     * Weigher function to use. If this is specified the cache turns into
     * a weighted cache and the function is called when cached data is stored
     * to determine its weight.
     */
    weigher?: Weigher<K, V> | null;
    /**
     * Listener to call whenever something is removed from the cache.
     */
    removalListener?: RemovalListener<K, V> | null;
}
/**
 * Bounded cache implementation using W-TinyLFU to keep track of data.
 *
 * See https://arxiv.org/pdf/1512.00727.pdf for details about TinyLFU and
 * the W-TinyLFU optimization.
 */
export declare class BoundedCache<K extends KeyType, V> extends AbstractCache<K, V> implements Cache<K, V>, CacheSPI<K, V> {
    private [DATA];
    [ON_REMOVE]?: RemovalListener<K, V>;
    [ON_EVICT]?: () => void;
    constructor(options: BoundedCacheOptions<K, V>);
    /**
     * Get the maximum size this cache can be.
     */
    get maxSize(): number;
    /**
     * Get the current size of the cache.
     */
    get size(): number;
    /**
     * Get the weighted size of all items in the cache.
     */
    get weightedSize(): number;
    /**
     * Cache and associate a value with the given key.
     */
    set(key: K, value: V): V | null;
    /**
     * Get a previously cached value.
     */
    getIfPresent(key: K): V | null;
    peek(key: K): V | null;
    /**
     * Delete any value associated with the given key from the cache.
     */
    delete(key: K): V | null;
    /**
     * Check if a certain value exists in the cache.
     */
    has(key: K): boolean;
    clear(): void;
    keys(): K[];
    cleanUp(): void;
    get metrics(): Metrics;
    private [TRIGGER_REMOVE];
    private [EVICT];
}
export {};
